import dataclasses
import logging
from contextlib import asynccontextmanager
from typing import AsyncGenerator

from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.pool import NullPool

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class DatabaseConnectionSettings:
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str
    POSTGRES_HOST: str
    POSTGRES_PORT: str
    POSTGRES_DB: str

    def get_dsn(self):
        dsn = f"postgresql://{self.POSTGRES_USER}:{self.POSTGRES_PASSWORD}@{self.POSTGRES_HOST}:{self.POSTGRES_PORT}/{self.POSTGRES_DB}"
        return dsn


class DatabaseConnectionManager:
    def __init__(self, settings: DatabaseConnectionSettings):
        self._engine = None
        self._async_sessionmaker = None
        self._settings = settings

    def _get_settings(self):
        if not self._settings:
            raise RuntimeError("No settings available")
        return self._settings

    def set_engine(self, engine: AsyncEngine):
        self._engine = engine

    def get_engine(self) -> AsyncEngine:
        if not self._engine:
            s = self._settings
            db_host = s.POSTGRES_HOST
            self._engine = create_async_engine(
                f"postgresql+asyncpg://{s.POSTGRES_USER}:{s.POSTGRES_PASSWORD}@{db_host}:{s.POSTGRES_PORT}/{s.POSTGRES_DB}",
                poolclass=NullPool,
            )
        return self._engine

    def get_session_maker(self):
        if not self._async_sessionmaker:
            self._async_sessionmaker = async_sessionmaker(
                self.get_engine(),
                expire_on_commit=False,
                class_=AsyncSession,
            )
        return self._async_sessionmaker

    @asynccontextmanager
    async def get_db_session(self) -> AsyncGenerator[AsyncSession, None]:
        session_maker = self.get_session_maker()
        async with session_maker() as session:
            try:
                yield session
                await session.commit()
            except Exception:
                await session.rollback()
                raise
            finally:
                await session.close()

    async def __call__(self):
        session_maker = self.get_session_maker()
        async with session_maker() as session:
            try:
                yield session
                await session.commit()
            except Exception:
                await session.rollback()
                raise
            finally:
                await session.close()


__all__ = ["DatabaseConnectionManager", "DatabaseConnectionSettings"]



import pytest

from toolbox.sqlalchemy.connection import DatabaseConnectionSettings


@pytest.fixture(autouse=True)
def db_settings():
    from pydantic import BaseModel
    data = dict(
        POSTGRES_USER="postgres",
        POSTGRES_PASSWORD = "postgres",
        POSTGRES_HOST = "0.0.0.0",
        POSTGRES_PORT = "5432",
        POSTGRES_DB = "postgres"
    )
    class TestSettings(BaseModel, DatabaseConnectionSettings):
        pass

    return TestSettings(**data)


@pytest.fixture(autouse=True)
async def temp_db(db_settings):
    from toolbox.testing import temporary_database
    from toolbox.sqlalchemy.models import BaseModel
    async with temporary_database(settings=db_settings, base_model=BaseModel):
        yield
        pass





from typing import Annotated

from fastapi import Depends
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession


from toolbox.schemes import SensitiveDataScheme, BaseScheme

AnyPydanticModel = Annotated[BaseScheme, Depends(BaseScheme)]
AnySQLAlchemyModel = Annotated[BaseScheme, Depends(SensitiveDataScheme)]


class AbstractDatabaseCrudManager:
    _alchemy_model = type[AnySQLAlchemyModel]
    _pydantic_model = type[AnyPydanticModel]

    @classmethod
    async def add_one(cls, conn: AsyncSession, data_model: AnyPydanticModel) -> None:
        obj = cls._alchemy_model(**data_model.encrypt_fields().model_dump())
        conn.add(obj)
        await conn.commit()
        await conn.refresh(obj)

    @classmethod
    async def get_all(cls, conn: AsyncSession, limit: int = 100, offset: int = 0):
        query = select(cls._alchemy_model).limit(limit).offset(offset)
        result = [v[0] for v in (await conn.execute(query)).all()]
        return result



from __future__ import annotations

import dataclasses
import logging
import time
from collections.abc import Callable
from typing import Iterable

from fastapi import Request
from fastapi import Response
from fastapi.security import HTTPBearer
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse
from starlette.types import ASGIApp

from toolbox.auth.enums import ResponseMessages

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class BearerTokenMiddlewareSettings:
    token_validator: Callable
    exclude_paths: Iterable[str]


class BearerTokenMiddleware(BaseHTTPMiddleware):
    _token_validator: Callable
    _exclude_paths: Iterable[str]

    def __init__(
        self,
        app,
    ):
        super().__init__(app)
        self.security = HTTPBearer()

    @classmethod
    def set_settings(cls, settings: BearerTokenMiddlewareSettings):
        cls._exclude_paths = settings.exclude_paths
        cls._token_validator = settings.token_validator

    async def _is_token_valid(self, token: str) -> bool:
        if await self._token_validator(token):
            return True
        logger.warning(
            "Token validation logic against external storage is not implemented. "
            "Access will be denied for all tokens."
        )
        return False

    async def dispatch(self, request: Request, call_next: Callable):
        if any(request.url.path.startswith(path) for path in self._exclude_paths):
            return await call_next(request)

        authorization = request.headers.get("Authorization")
        if not authorization:
            return JSONResponse(
                status_code=401,
                content=ResponseMessages.header_missing,
            )

        scheme, token = authorization.split()
        if scheme.lower() != "bearer":
            return JSONResponse(
                status_code=401,
                content=ResponseMessages.invalid_scheme,
            )

        if not await self._is_token_valid(token):
            return JSONResponse(
                status_code=403,
                content=ResponseMessages.invalid_token,
            )

        response = await call_next(request)
        return response


class LoggerMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp):
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        start_time = time.time()

        path_components = request.url.path.split("/")
        bot_id = path_components[-1] if len(path_components) > 1 else "unknown"

        logger.info(
            f""""{request.method} {request.url.path}" - {bot_id} {request.client.host if request.client else 'unknown'}"""
        )

        try:
            response = await call_next(request)
            process_time = time.time() - start_time

            logger.info(
                f""""{request.method} {request.url.path}" {response.status_code} - {process_time:.3f}s - {bot_id} """
            )
            return response
        except Exception as e:
            process_time = time.time() - start_time
            logger.error(f""""{request.method} {request.url.path}" 500 - {process_time:.3f}s - {bot_id} | {str(e)} """)
            return JSONResponse(status_code=500, content=ResponseMessages.internal_server_error)



from unittest.mock import MagicMock, patch

import pytest
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from tests.fixtures.database import db_settings, temp_db
from tests.fixtures.db_connect import database_connector


@pytest.mark.asyncio
async def test_database_manager_no_settings(database_connector):
    database_connector._settings = None
    database_connector._engine = None
    database_connector._async_sessionmaker = None

    with pytest.raises(RuntimeError, match="No settings available"):
        database_connector._get_settings()


@patch('toolbox.sqlalchemy.connection.create_async_engine')
def test_database_manager_get_engine_creates_new(mock_create_engine, database_connector, db_settings):
    database_connector._engine = None

    expected_url = f"postgresql+asyncpg://{db_settings.POSTGRES_USER}:{db_settings.POSTGRES_PASSWORD}@{db_settings.POSTGRES_HOST}:{db_settings.POSTGRES_PORT}/{db_settings.POSTGRES_DB}"

    database_connector.get_engine()

    mock_create_engine.assert_called_once()
    call_args = mock_create_engine.call_args[0][0]
    assert call_args == expected_url


def test_database_manager_get_session_maker_returns_existing(database_connector):
    mock_sessionmaker = MagicMock()
    database_connector._async_sessionmaker = mock_sessionmaker

    result = database_connector.get_session_maker()
    assert result == mock_sessionmaker


async def test_fastapi_depends_itegration_test(temp_db, database_connector):
    from fastapi import Depends, FastAPI
    app = FastAPI()
    @app.get("/")
    async def index(database_conn = Depends(database_connector)):
        async with database_conn:
            res = await database_conn.scalar(select(1))
        return {"status": "ok"}

    from httpx import ASGITransport, AsyncClient
    async with AsyncClient(
            transport=ASGITransport(app=app), base_url=f"http://test",
    ) as client:
        response1 = await client.get('/')
        assert response1.status_code == 200


async def test_get_db_connect_context_works(temp_db, database_connector):
    async with database_connector.get_db_session() as conn:
        res = await conn.scalar(select(1))



from sqlalchemy import orm


def lenient_constructor(self, **kwargs):
    cls_ = type(self)
    for k in kwargs:
        if hasattr(cls_, k):
            setattr(self, k, kwargs[k])


registry = orm.registry(constructor=lenient_constructor)


class BaseModel(orm.DeclarativeBase):
    registry = registry



import pytest
from fastapi import FastAPI
from httpx import ASGITransport
from httpx import AsyncClient

from toolbox.auth.enums import ResponseMessages


async def token_validator(self, token: str) -> bool:
    if token == "valid_token_321":
        return True
    return False


@pytest.fixture(autouse=True)
def config():
    pass

@pytest.fixture(autouse=True)
def app(config) -> FastAPI:
    app_ = FastAPI()
    from toolbox.auth.middlewares.fastapi_ import BearerTokenMiddleware, BearerTokenMiddlewareSettings
    s = BearerTokenMiddlewareSettings(
        token_validator=token_validator,
        exclude_paths=[]
    )
    BearerTokenMiddleware.set_settings(s)
    app_.add_middleware(BearerTokenMiddleware)

    @app_.get('/auth')
    def auth():
        return {'auth': 'success'}

    return app_


async def test_auth_middleware_403_invalid_token(app: FastAPI, config):
    async with AsyncClient(
            transport=ASGITransport(app=app), base_url='http://test',
            headers={'Authorization': 'Bearer invalid_token_123'},
    ) as client:
        response = await client.get('/auth')

    assert response.status_code == 403
    current_responce = response.json()
    assert current_responce == ResponseMessages.invalid_token


async def test_auth_middleware_401_invalid_scheme(app: FastAPI, config):
    async with AsyncClient(
            transport=ASGITransport(app=app), base_url='http://test', headers={'Authorization': 'NotBearer some_token'},
    ) as client:
        response = await client.get('/auth')

    assert response.status_code == 401
    current_responce = response.json()
    assert current_responce == ResponseMessages.invalid_scheme


async def test_auth_middleware_401_header_missing(app: FastAPI, config):
    async with AsyncClient(
            transport=ASGITransport(app=app), base_url='http://test', headers={},
    ) as client:
        response = await client.get('/auth')

    assert response.status_code == 401
    current_responce = response.json()
    assert current_responce == ResponseMessages.header_missing


async def test_auth_middleware_200(app: FastAPI, config):
    async with AsyncClient(
            transport=ASGITransport(app=app), base_url='http://test',
            headers={'Authorization': f'Bearer valid_token_321'},
    ) as client:
        response = await client.get('/auth')

    assert response.status_code == 200
    current_responce = response.json()
    assert current_responce == {'auth': 'success'}



import asyncio


def async_to_sync(awaitable):
    """Run your async code in sync mode"""
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(awaitable)


def sync_to_async(func):
    """Run your sync code in async mode"""

    async def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

from enum import Enum


class ResponseMessages(str, Enum):
    header_missing = "Authorization header missing"
    invalid_scheme = "Invalid authentication scheme"
    invalid_token = "Invalid token or insufficient permissions"
    internal_server_error = "Internal Server Error"



import logging
from contextlib import asynccontextmanager

import asyncpg

logger = logging.getLogger(__name__)


@asynccontextmanager
async def debug_client(app, app_path: str = "http://test"):
    from httpx import ASGITransport, AsyncClient

    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url=app_path,
    ) as client:
        yield client
        pass


@asynccontextmanager
async def temporary_database(settings: "DatabaseConnectionSettings", base_model, db_prefix: str = "test"):
    original_settings = settings.__class__(**settings.__dict__)
    test_db_name = f"{db_prefix}_{original_settings.POSTGRES_DB}"
    settings.POSTGRES_DB = test_db_name

    from toolbox.sqlalchemy.connection import DatabaseConnectionManager

    connection_factory = DatabaseConnectionManager(settings=original_settings)

    dsn = settings.get_dsn()
    try:
        conn = await asyncpg.connect(dsn=dsn)
    except:
        conn = await asyncpg.connect(dsn=dsn.replace(f"/{settings.POSTGRES_DB}", "/postgres"))
        await conn.execute(f"CREATE DATABASE {settings.POSTGRES_DB}")

    async_engine = connection_factory.get_engine()
    async with async_engine.begin() as conn:
        await conn.run_sync(base_model.metadata.create_all)
        logger.debug({"msg": "Migration 'settings.POSTGRES_DB' was completed"})
        yield

        await conn.run_sync(base_model.metadata.drop_all)



import pytest

from toolbox.sqlalchemy.connection import DatabaseConnectionManager
from tests.fixtures.database import temp_db, db_settings


@pytest.fixture(autouse=True)
def database_connector(temp_db, db_settings):
    dc = DatabaseConnectionManager(settings=db_settings)
    return dc



from sqlalchemy import select

from tests.fixtures.database import temp_db, db_settings
from tests.fixtures.db_connect import database_connector


async def test_fastapi_depends_itegration_test(temp_db, db_settings, database_connector):

    from fastapi import Depends, FastAPI
    app = FastAPI()
    @app.get("/")
    async def index(database_conn = Depends(database_connector)):
        await database_conn.scalar(select(1))
        return {"status": "ok"}

    from toolbox.testing import debug_client
    async with debug_client(app) as client:
        response1 = await client.get('/')
        assert response1.status_code == 200


async def test_debug_client_negative_test(temp_db, db_settings, database_connector):

    from fastapi import FastAPI
    app = FastAPI()

    from toolbox.testing import debug_client
    async with debug_client(app) as client:
        response1 = await client.get('/')
        assert response1.status_code == 404



import pytest
from unittest.mock import MagicMock, AsyncMock, patch
from sqlalchemy import select, String
from sqlalchemy.orm import mapped_column, Mapped
from sqlalchemy.sql.selectable import Select

from toolbox.sqlalchemy.repositories import AbstractDatabaseCrudManager
from toolbox.sqlalchemy.models import BaseModel
from toolbox.schemes import BaseScheme, SensitiveDataScheme


class TestPydanticModel(BaseScheme):
    name: str
    value: str

    def encrypt_fields(self):
        return self


class TestSQLAlchemyModel(BaseModel):
    __tablename__ = "test_table"
    name: Mapped[str] = mapped_column(String, primary_key=True)
    value: Mapped[str] = mapped_column(String)


class TestCrudManager(AbstractDatabaseCrudManager):
    _alchemy_model = TestSQLAlchemyModel
    _pydantic_model = TestPydanticModel


@pytest.fixture
def test_data():
    return TestPydanticModel(name="test_name", value="test_value")


@pytest.fixture
def mock_session():
    session = AsyncMock()

    async_context_manager_mock = AsyncMock()

    async_context_manager_mock.__aenter__.return_value = session

    async_context_manager_mock.__aexit__.return_value = False

    session_maker = MagicMock()
    session_maker.return_value = async_context_manager_mock

    return session_maker, session


@pytest.mark.asyncio
@patch('toolbox.sqlalchemy.connection.DatabaseConnectionManager')
async def test_add_one(mock_db_manager, mock_session, test_data):
    session_maker, session = mock_session
    await TestCrudManager.add_one(session, test_data)

    session.add.assert_called_once()
    added_model = session.add.call_args[0][0]
    assert isinstance(added_model, TestSQLAlchemyModel)
    assert added_model.name == test_data.name
    assert added_model.value == test_data.value

    session.commit.assert_called_once()
    session.refresh.assert_called_once_with(added_model)


@pytest.mark.asyncio
@patch('toolbox.sqlalchemy.connection.DatabaseConnectionManager')
async def test_get_all(mock_db_manager, mock_session):
    session_maker, session = mock_session

    test_models = [
        TestSQLAlchemyModel(name="test1", value="value1"),
        TestSQLAlchemyModel(name="test2", value="value2")
    ]

    db_operation_result_mock = MagicMock()
    db_operation_result_mock.all.return_value = [(model,) for model in test_models]

    session.execute.return_value = db_operation_result_mock

    result = await TestCrudManager.get_all(session, limit=10, offset=0)

    session.execute.assert_called_once()
    query = session.execute.call_args[0][0]
    assert isinstance(query, Select)

    assert len(result) == 2
    assert all(isinstance(item, TestSQLAlchemyModel) for item in result)
    assert result == test_models


@pytest.mark.asyncio
@patch('toolbox.sqlalchemy.connection.DatabaseConnectionManager')
async def test_get_all_with_custom_limit_offset(mock_db_manager, mock_session):
    session_maker, session = mock_session

    db_operation_result_mock = MagicMock()
    db_operation_result_mock.all.return_value = []
    session.execute.return_value = db_operation_result_mock

    await TestCrudManager.get_all(session, limit=5, offset=10)

    session.execute.assert_called_once()
    query = session.execute.call_args[0][0]

    assert query._limit_clause is not None
    assert query._limit_clause.value == 5
    assert query._offset_clause is not None
    assert query._offset_clause.value == 10


@pytest.mark.asyncio
@patch('toolbox.sqlalchemy.connection.DatabaseConnectionManager')
async def test_get_all_empty_result(mock_db_manager, mock_session):
    session_maker, session = mock_session

    execute_result = MagicMock()
    execute_result.all.return_value = []
    session.execute.return_value = execute_result

    result = await TestCrudManager.get_all(session)

    assert result == []
    session.execute.assert_called_once()
